@@ -1,8 +1,8 @@
 import numpy as np
-import numba
-from numba import jit
 import h5py
 from matplotlib import pyplot as plt
+import torch
+from tqdm import tqdm
 
 from mrftools import DictionaryParameters, SequenceParameters
 
@@ -25,62 +25,63 @@ class Simulation:
             self.name = name
 
     @staticmethod
-    @jit(parallel=True, nopython=True)
-    def ExecuteNumbaSimulation(numTimepoints,T1s,T2s,B1s,TRs,TEs,FAs,phaseValues):
+    def Simulate(numTimepoints,T1s,T2s,B1s,TRs,TEs,FAs,spinOffresonances,device=None):     
+        if(device==None):
+            if torch.cuda.is_available():
+                device = torch.device("cuda")
+            else:
+                device = torch.device("cpu")    
 
-        numberOfExperiments = np.int32(numTimepoints)         
-        numberOfDictionaryEntries = len(T1s)
-        numberOfSpins = len(phaseValues)
-        FAs = np.radians(FAs)
-        Mx0 = np.zeros((numTimepoints,numberOfDictionaryEntries,numberOfSpins))
-        My0 = np.zeros((numberOfExperiments,numberOfDictionaryEntries,numberOfSpins))
-        Mz0 = np.zeros((numberOfExperiments,numberOfDictionaryEntries,numberOfSpins))
-        phaseValueCosines = np.cos(phaseValues)
-        phaseValueSines = np.sin(phaseValues)
-                            
-        for dictionaryEntryNumber in numba.prange(0,numberOfDictionaryEntries):
-            
-            T1 = T1s[dictionaryEntryNumber]
-            T2 = T2s[dictionaryEntryNumber]
-            B1 = B1s[dictionaryEntryNumber]
-            
-            Mx = np.zeros(numberOfSpins)
-            My = np.zeros(numberOfSpins)
-            Mz = np.zeros(numberOfSpins)
-            Mz[:] = -0.95
-            
+        T1s = torch.tensor(T1s).to(device)
+        T2s = torch.tensor(T2s).to(device)
+        B1s = torch.tensor(B1s).to(device)
+        FAs = torch.tensor(FAs).to(device)
+        spinOffresonances = torch.tensor(spinOffresonances).to(device)
+        Mx0 = torch.zeros((numTimepoints,len(spinOffresonances),len(T1s)))
+        My0 = torch.zeros((numTimepoints,len(spinOffresonances),len(T1s)))
+        Mz0 = torch.zeros((numTimepoints,len(spinOffresonances),len(T1s)))
+        Mx = torch.zeros((len(spinOffresonances),len(T1s))).to(device)
+        My = torch.zeros((len(spinOffresonances),len(T1s))).to(device)
+        Mz = torch.zeros((len(spinOffresonances),len(T1s))).to(device)
+
+        FAs = torch.deg2rad(FAs)
+        phaseValueCosines = torch.cos(spinOffresonances)
+        phaseValueSines = torch.sin(spinOffresonances)
+        
+        Mz[:] = -0.95
+        with torch.no_grad():
             for iTimepoint in range(numTimepoints):
-                fa = FAs[iTimepoint] * B1
+                fa = FAs[iTimepoint] * B1s
                 tr = TRs[iTimepoint]
                 te = TEs[iTimepoint]
                 tre = tr-te
                 
-                At2te = np.exp(-1*te/T2)
-                At1te = np.exp(-1*te/T1)
+                At2te = torch.exp(-1*te/T2s)
+                At1te = torch.exp(-1*te/T1s)
                 Bt1te = 1-At1te
                 
-                At2tr = np.exp(-1*tre/T2)
-                At1tr = np.exp(-1*tre/T1)
+                At2tr = torch.exp(-1*tre/T2s)
+                At1tr = torch.exp(-1*tre/T1s)
                 Bt1tr = 1-At1tr
             
-                crf = np.cos(fa)
-                srf = np.sin(fa)
+                crf = torch.cos(fa)
+                srf = torch.sin(fa)
 
                 # Applying flip angle 
                 Myi = My
                 Mzi = Mz
-                My = np.multiply(crf,Myi)-np.multiply(srf,Mzi)
-                Mz = np.multiply(srf,Myi)+np.multiply(crf,Mzi)
+                My = torch.multiply(crf,Myi)-torch.multiply(srf,Mzi)
+                Mz = torch.multiply(srf,Myi)+torch.multiply(crf,Mzi)
 
                 # Relaxation over TE
-                Mx = np.multiply(Mx,At2te)
-                My = np.multiply(My,At2te)
-                Mz = np.multiply(Mz,At1te)+Bt1te
+                Mx = torch.multiply(Mx, At2te)
+                My = torch.multiply(My, At2te)
+                Mz = torch.multiply(Mz, At1te)+Bt1te
 
-                # Reading value after TE and before TRE
-                Mx0[iTimepoint,dictionaryEntryNumber,:]=Mx[:]
-                My0[iTimepoint,dictionaryEntryNumber,:]=My[:]
-                Mz0[iTimepoint,dictionaryEntryNumber,:]=Mz[:]
+                # Reading value after TE and before TRE 
+                Mx0[iTimepoint,:,:]=Mx.cpu()
+                My0[iTimepoint,:,:]=My.cpu()
+                Mz0[iTimepoint,:,:]=Mz.cpu()
 
                 # Relaxation over TRE (TR-TE) 
                 Mx = Mx*At2tr
@@ -88,60 +89,43 @@ class Simulation:
                 Mz = Mz*At1tr+Bt1tr
 
                 # Applying off-resonance to spins
-                Mxi = Mx
-                Myi = My
-                Mx = np.multiply(phaseValueCosines,Mxi) - np.multiply(phaseValueSines,Myi)
-                My = np.multiply(phaseValueSines,Mxi) + np.multiply(phaseValueCosines,Myi)
-
+                Mxi = Mx.t()
+                Myi = My.t()
+                Mx = (torch.multiply(phaseValueCosines,Mxi) - torch.multiply(phaseValueSines,Myi)).t()
+                My = (torch.multiply(phaseValueSines,Mxi) + torch.multiply(phaseValueCosines,Myi)).t()
+                del fa, tr, te, tre, At2te, At1te, Bt1te, At2tr, At1tr, Bt1tr, crf, srf, Mxi, Myi, Mzi
+        del T1s, T2s, B1s, FAs, spinOffresonances, Mx, My, Mz
         return Mx0,My0,Mz0
-        
-    @staticmethod
-    @jit(parallel=True, nopython=True)
-    def ParallizedMeans(Mx0,My0,Mz0):
-        MeansXo = np.zeros((np.shape(Mx0)[0],np.shape(Mx0)[1]))
-        MeansYo = np.zeros((np.shape(Mx0)[0],np.shape(Mx0)[1]))
-        MeansZo = np.zeros((np.shape(Mx0)[0],np.shape(Mx0)[1]))
-
-        for n in numba.prange(np.shape(Mx0)[1]):
-            for timepoint in range(np.shape(Mx0)[0]):
-                MeansXo[timepoint,n] = np.mean(Mx0[timepoint,n,:])
-                MeansYo[timepoint,n] = np.mean(My0[timepoint,n,:])
-                MeansZo[timepoint,n] = np.mean(Mz0[timepoint,n,:])
-        return MeansXo, MeansYo, MeansZo
-
-    def Execute(self, numBatches = 1):
-        TRs = []
-        TEs = []
-        FAs = []
-        T1s= []
-        T2s = []
-        B1s = []
-
-        for timepoint in self.sequenceParameters.timepoints:
-            TRs.append(timepoint['TR'])
-            TEs.append(timepoint['TE'])
-            FAs.append(timepoint['FA'])
 
+    def Execute(self, numBatches=1, device=None):
+        if(device==None):
+            if torch.cuda.is_available():
+                device = torch.device("cuda")
+            else:
+                device = torch.device("cpu")
+        TRs = np.copy(self.sequenceParameters.timepoints['TR'])
+        TEs = np.copy(self.sequenceParameters.timepoints['TE'])
+        FAs = np.copy(self.sequenceParameters.timepoints['FA'])
         phaseValues = np.linspace(self.phaseRange[0], self.phaseRange[1], self.numSpins)
         dictEntriesPerBatch = int(len(self.dictionaryParameters.entries)/numBatches)
         print("Simulating " + str(numBatches) + " batch(s) of ~" + str(dictEntriesPerBatch) + " dictionary entries")
         Mxy = np.zeros((self.numTimepoints, len(self.dictionaryParameters.entries)), np.complex128)
-        for i in range(numBatches):
-            firstDictEntry = i*dictEntriesPerBatch
-            if i == (numBatches-1):
-                lastDictEntry = len(self.dictionaryParameters.entries)
-            else:
-                lastDictEntry = firstDictEntry+dictEntriesPerBatch
-            print("Starting Batch: " + str(i) + " (Entries " + str(firstDictEntry) + "->" + str(lastDictEntry)+ ")")
-            T1s.clear();T2s.clear();B1s.clear()
-            for dictionaryEntry in self.dictionaryParameters.entries[firstDictEntry:lastDictEntry]:
-                T1s.append(dictionaryEntry['T1'])
-                T2s.append(dictionaryEntry['T2'])
-                B1s.append(dictionaryEntry['B1'])
-            Mx0,My0,Mz0 = self.ExecuteNumbaSimulation(self.numTimepoints, np.asarray(T1s),np.asarray(T2s),np.asarray(B1s),np.asarray(TRs),np.asarray(TEs),np.asarray(FAs),phaseValues)
-            Mx,My,Mz = self.ParallizedMeans(Mx0,My0,Mz0)
-            Mxy[:,firstDictEntry:lastDictEntry] = Mx+(My*1j)
-        self.results = Mxy
+        with tqdm(total=numBatches) as pbar:
+            for i in range(numBatches):
+                firstDictEntry = i*dictEntriesPerBatch
+                if i == (numBatches-1):
+                    lastDictEntry = len(self.dictionaryParameters.entries)
+                else:
+                    lastDictEntry = firstDictEntry+dictEntriesPerBatch
+                T1s = np.copy(self.dictionaryParameters.entries[firstDictEntry:lastDictEntry]['T1'])
+                T2s = np.copy(self.dictionaryParameters.entries[firstDictEntry:lastDictEntry]['T2'])
+                B1s = np.copy(self.dictionaryParameters.entries[firstDictEntry:lastDictEntry]['B1'])
+                Mx0,My0,Mz0 = self.Simulate(self.numTimepoints, T1s, T2s, B1s, TRs, TEs, FAs, phaseValues, device=device)
+                Mx = torch.mean(Mx0, axis=1)    
+                My = torch.mean(My0, axis=1)    
+                Mxy[:,firstDictEntry:lastDictEntry] = Mx+(My*1j)
+                pbar.update(1)
+            self.results = Mxy
         return self.results
 
     def CalculateSVD(self, desiredSVDPower=0.99, truncationNumberOverride=None):
